/*
 * Photos - access, organize and share your photos on GNOME
 * Copyright © 2012 – 2019 Red Hat, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Based on code from:
 *   + Documents
 */


#include "config.h"

#include <string.h>

#include "photos-application.h"
#include "photos-base-manager.h"
#include "photos-query.h"
#include "photos-query-builder.h"
#include "photos-search-type.h"
#include "photos-source-manager.h"
#include "photos-search-match-manager.h"
#include "photos-search-type-manager.h"
#include "photos-tracker-queue.h"

#define PHOTOS_QUERY_COLLECTIONS_IDENTIFIER "photos:collection:"
#define PHOTOS_QUERY_LOCAL_COLLECTIONS_IDENTIFIER "photos:collection:local:"

const gchar *collections_default_filter = \
  "(fn:starts-with (nao:identifier (?urn), '" PHOTOS_QUERY_COLLECTIONS_IDENTIFIER "')"
  "   || (?urn = nfo:image-category-screenshot))";


/* This includes mimetype blocklist */
const gchar *photos_default_filter = \
  "(nie:mimeType(?urn) != 'image/gif' && nie:mimeType(?urn) != 'image/x-eps')";


static gchar *
photos_query_builder_query (PhotosSearchContextState *state,
                            const gchar *values,
                            gint flags,
                            PhotosOffsetController *offset_cntrlr)
{
  PhotosSparqlTemplate *template;
  const gchar *miner_fs_busname = NULL;
  const gchar *main_projection = NULL;
  const gchar *second_projection = NULL;
  g_autofree gchar *item_pattern = NULL;
  g_autofree gchar *search_filter = NULL;
  g_autofree gchar *source_filter = NULL;
  const gchar *order = NULL;
  g_autofree gchar *offset_limit = NULL;
  gchar *sparql;

  template = photos_base_manager_get_sparql_template (state->srch_typ_mngr, flags);

  miner_fs_busname = photos_tracker_queue_get_miner_fs_busname (state->queue);

  main_projection = "?urn "
                    "?file "
                    "nfo:fileName (?file) AS ?filename "
                    "nie:mimeType (?urn) AS ?mimetype "
                    "nie:title (?urn) AS ?title "
                    "tracker:coalesce (nco:fullname (?creator), nco:fullname (?publisher), '') AS ?author_name "
                    "tracker:coalesce (nfo:fileLastModified (?file), nie:contentLastModified (?urn)) AS ?mtime "
                    "nao:identifier (?urn) AS ?identifier "
                    "rdf:type (?urn) AS ?type "
                    "nie:dataSource(?urn) AS ?datasource "
                    "( EXISTS { ?urn nco:contributor ?contributor FILTER ( ?contributor != ?creator ) } ) AS ?has_contributor "
                    "tracker:coalesce(nfo:fileCreated (?file), nie:contentCreated (?urn)) AS ?ctime "
                    "nfo:width (?urn) AS ?width "
                    "nfo:height (?urn) AS ?height "
                    "nfo:equipment (?urn) AS ?equipment "
                    "nfo:orientation (?urn) AS ?orientation "
                    "nmm:exposureTime (?urn) AS ?exposure_time "
                    "nmm:fnumber (?urn) AS ?fnumber "
                    "nmm:focalLength (?urn) AS ?focal_length "
                    "nmm:isoSpeed (?urn) AS ?isospeed "
                    "nmm:flash (?urn) AS ?flash "
                    "slo:location (?urn) AS ?location ";

  second_projection = "?urn ?file ?filename ?mimetype ?title ?author_name ?mtime ?identifier ?type ?datasource "
                      "( EXISTS { ?urn nao:hasTag nao:predefined-tag-favorite } ) AS ?is_favorite "
                      "?has_contributor ?ctime ?width ?height ?equipment ?orientation ?exposure_time ?fnumber "
                      "?focal_length ?isospeed ?flash ?location ";

  item_pattern = photos_base_manager_get_where (state->item_mngr, flags);

  if (!(flags & PHOTOS_QUERY_FLAGS_UNFILTERED))
    {
      source_filter = photos_base_manager_get_filter (state->src_mngr, flags);
      search_filter = photos_base_manager_get_filter (state->srch_mtch_mngr, flags);
    }

  order = "ORDER BY DESC (?mtime)";

  if (values == NULL && (flags & PHOTOS_QUERY_FLAGS_UNLIMITED) == 0)
    {
      gint offset = 0;
      gint step = 60;

      if (offset_cntrlr != NULL)
        {
          offset = photos_offset_controller_get_offset (offset_cntrlr);
          step = photos_offset_controller_get_step (offset_cntrlr);
        }

      offset_limit = g_strdup_printf ("LIMIT %d OFFSET %d", step, offset);
    }

  sparql = photos_sparql_template_get_sparql (template,
                                              "miner_fs_busname", miner_fs_busname,
                                              "main_projection", main_projection,
                                              "second_projection", second_projection,
                                              "final_projection", second_projection,
                                              "values", values ? values : "",
                                              "collections_default_filter", collections_default_filter,
                                              "item_pattern", item_pattern,
                                              "photos_default_filter", photos_default_filter,
                                              "source_filter", source_filter ? source_filter : "(true)",
                                              "search_filter", search_filter ? search_filter : "(true)",
                                              "order", order,
                                              "offset_limit", offset_limit ? offset_limit : "",
                                              NULL);

  return sparql;
}


PhotosQuery *
photos_query_builder_create_collection_query (PhotosSearchContextState *state,
                                              const gchar *name,
                                              const gchar *identifier_tag)
{
  GTimeVal tv;
  PhotosQuery *query;
  g_autoptr (TrackerResource) collection = NULL;
  g_autofree gchar *identifier = NULL;
  g_autofree gchar *sparql = NULL;
  g_autofree gchar *time = NULL;
  gint64 timestamp;

  identifier = g_strdup_printf ("%s%s",
                                PHOTOS_QUERY_LOCAL_COLLECTIONS_IDENTIFIER,
                                identifier_tag == NULL ? name : identifier_tag);

  timestamp = g_get_real_time () / G_USEC_PER_SEC;
  tv.tv_sec = timestamp;
  tv.tv_usec = 0;
  time = g_time_val_to_iso8601 (&tv);

  collection = tracker_resource_new ("_:res");
  tracker_resource_add_uri (collection, "rdf:type", "nfo:DataContainer");
  tracker_resource_add_uri (collection, "rdf:type", "nie:DataObject");
  tracker_resource_set_string (collection, "nie:contentLastModified", time);
  tracker_resource_set_string (collection, "nie:title", name);
  tracker_resource_set_string (collection, "nao:identifier", identifier);

  sparql = tracker_resource_print_sparql_update (collection, NULL, "tracker:Pictures");

  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_collection_icon_query (PhotosSearchContextState *state, const gchar *resource)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;

  sparql = g_strdup_printf ("SELECT ?urn "
                            "tracker:coalesce(nfo:fileLastModified(?urn), nie:contentLastModified(?urn)) AS ?mtime "
                            "WHERE { ?urn nie:isPartOf <%s> } "
                            "ORDER BY DESC (?mtime) LIMIT 4",
                            resource);

  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_count_query (PhotosSearchContextState *state, gint flags)
{
  PhotosSparqlTemplate *template;
  const gchar *miner_fs_busname = NULL;
  const gchar *count_projection = NULL;
  const gchar *value_projection = NULL;
  g_autofree gchar *item_pattern = NULL;
  g_autofree gchar *search_filter = NULL;
  g_autofree gchar *source_filter = NULL;
  g_autofree gchar *sparql = NULL;
  PhotosQuery *query;

  template = photos_base_manager_get_sparql_template (state->srch_typ_mngr, flags);

  miner_fs_busname = photos_tracker_queue_get_miner_fs_busname (state->queue);

  value_projection = "?urn ";
  count_projection = "COUNT(?urn) ";

  item_pattern = photos_base_manager_get_where (state->item_mngr, flags);

  if (! (flags & PHOTOS_QUERY_FLAGS_UNFILTERED))
    {
      source_filter = photos_base_manager_get_filter (state->src_mngr, flags);
      search_filter = photos_base_manager_get_filter (state->srch_mtch_mngr, flags);
    }

  sparql = photos_sparql_template_get_sparql (template,
                                              "miner_fs_busname", miner_fs_busname,
                                              "main_projection", value_projection,
                                              "second_projection", value_projection,
                                              "final_projection", count_projection,
                                              "values", "",
                                              "collections_default_filter", collections_default_filter,
                                              "item_pattern", item_pattern,
                                              "photos_default_filter", photos_default_filter,
                                              "source_filter", source_filter ? source_filter : "(true)",
                                              "search_filter", search_filter ? search_filter : "(true)",
                                              "order", "",
                                              "offset_limit", "",
                                              NULL);

  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_delete_resource_query (PhotosSearchContextState *state, const gchar *resource)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;

  sparql = g_strdup_printf ("DELETE { <%s> a rdfs:Resource }", resource);
  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_equipment_query (PhotosSearchContextState *state, GQuark equipment)
{
  PhotosQuery *query;
  const gchar *resource;
  g_autofree gchar *sparql = NULL;

  resource = g_quark_to_string (equipment);
  sparql = g_strdup_printf ("SELECT nfo:manufacturer (<%s>) nfo:model (<%s>) WHERE {}", resource, resource);
  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_fetch_collections_for_urn_query (PhotosSearchContextState *state, const gchar *resource)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;

  sparql = g_strdup_printf ("SELECT ?urn WHERE { ?urn a nfo:DataContainer . <%s> nie:isPartOf ?urn }", resource);
  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_fetch_collections_local (PhotosSearchContextState *state)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;

  sparql = photos_query_builder_query (state,
                                       NULL,
                                       PHOTOS_QUERY_FLAGS_COLLECTIONS
                                       | PHOTOS_QUERY_FLAGS_LOCAL
                                       | PHOTOS_QUERY_FLAGS_UNLIMITED,
                                       NULL);

  query = photos_query_new (NULL, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_global_query (PhotosSearchContextState *state,
                                   gint flags,
                                   PhotosOffsetController *offset_cntrlr)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;

  sparql = photos_query_builder_query (state, NULL, flags, offset_cntrlr);
  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_location_query (PhotosSearchContextState *state, const gchar *location_urn)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;

  sparql = g_strdup_printf ("SELECT slo:latitude (<%s>) slo:longitude (<%s>) WHERE {}", location_urn, location_urn);
  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_set_collection_query (PhotosSearchContextState *state,
                                           const gchar *item_urn,
                                           const gchar *collection_urn,
                                           gboolean setting)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;

  if (setting)
    sparql = g_strdup_printf ("INSERT DATA { "
                              "  GRAPH tracker:Pictures {"
                              "    <%s> a nie:DataObject , nmm:Photo ; "
                              "      nie:isPartOf <%s> "
                              "  }"
                              "}", item_urn, collection_urn);
  else
    sparql = g_strdup_printf ("DELETE DATA { "
                              "  GRAPH tracker:Pictures {"
                              "    <%s> nie:isPartOf <%s> "
                              "  }"
                              "}", item_urn, collection_urn);
  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_single_query (PhotosSearchContextState *state, gint flags, const gchar *resource)
{
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;
  g_autofree gchar *values = NULL;

  values = g_strdup_printf ("VALUES ?urn { <%s> }", resource);

  sparql = photos_query_builder_query (state, values, flags, NULL);

  query = photos_query_new (state, sparql);

  return query;
}


PhotosQuery *
photos_query_builder_update_mtime_query (PhotosSearchContextState *state, const gchar *resource)
{
  GTimeVal tv;
  PhotosQuery *query;
  g_autofree gchar *sparql = NULL;
  g_autofree gchar *time = NULL;
  gint64 timestamp;

  timestamp = g_get_real_time () / G_USEC_PER_SEC;
  tv.tv_sec = timestamp;
  tv.tv_usec = 0;
  time = g_time_val_to_iso8601 (&tv);

  sparql = g_strdup_printf ("WITH tracker:Pictures "
                            "DELETE { <%s> nie:contentLastModified ?time } "
                            "INSERT { "
                            "  <%s> a nmm:Photo ; "
                            "    nie:contentLastModified '%s' "
                            "}"
                            "WHERE { <%s> nie:contentLastModified ?time }",
                            resource, resource, time, resource);
  query = photos_query_new (state, sparql);

  return query;
}
