/*
 * Photos - access, organize and share your photos on GNOME
 * Copyright © 2012 – 2019 Red Hat, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Based on code from:
 *   + Documents
 */


#include "config.h"

#include <glib.h>
#include <tracker-sparql.h>

#include "photos-debug.h"
#include "photos-filterable.h"
#include "photos-source.h"
#include "photos-tracker-queue.h"


struct _PhotosTrackerQueue
{
  GObject parent_instance;
  GError *initialization_error;
  GQueue *queue;
  TrackerSparqlConnection *local_connection;
  TrackerSparqlConnection *miner_fs_connection;
  gboolean is_initialized;
  gboolean running;
  gboolean miner_fs_ready;
  const gchar *miner_fs_busname;
  const gchar *miner_fs_control_busname;
};

enum
{
  PROP_0,
  PROP_MINER_FS_READY,
  PROP_MINER_FS_BUSNAME
};

static void photos_tracker_queue_initable_iface_init (GInitableIface *iface);


G_DEFINE_TYPE_WITH_CODE (PhotosTrackerQueue, photos_tracker_queue, G_TYPE_OBJECT,
                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, photos_tracker_queue_initable_iface_init));


typedef enum
{
  PHOTOS_TRACKER_QUERY_SELECT,
  PHOTOS_TRACKER_QUERY_UPDATE,
  PHOTOS_TRACKER_QUERY_UPDATE_BLANK
} PhotosTrackerQueryType;

typedef struct _PhotosTrackerQueueData PhotosTrackerQueueData;

struct _PhotosTrackerQueueData
{
  GAsyncReadyCallback callback;
  GCancellable *cancellable;
  GDestroyNotify destroy_data;
  PhotosQuery *query;
  PhotosTrackerQueryType query_type;
  gpointer user_data;
};


G_LOCK_DEFINE_STATIC (init_lock);


static void photos_tracker_queue_check (PhotosTrackerQueue *self);


static void
photos_tracker_queue_data_free (PhotosTrackerQueueData *data)
{
  g_clear_object (&data->query);
  g_clear_object (&data->cancellable);

  if (data->destroy_data != NULL)
    (*data->destroy_data) (data->user_data);

  g_slice_free (PhotosTrackerQueueData, data);
}


G_DEFINE_AUTOPTR_CLEANUP_FUNC (PhotosTrackerQueueData, photos_tracker_queue_data_free);

static PhotosTrackerQueueData *
photos_tracker_queue_data_new (PhotosQuery *query,
                               PhotosTrackerQueryType query_type,
                               GCancellable *cancellable,
                               GAsyncReadyCallback callback,
                               gpointer user_data,
                               GDestroyNotify destroy_data)
{
  PhotosTrackerQueueData *data;

  data = g_slice_new0 (PhotosTrackerQueueData);

  data->query = g_object_ref (query);
  data->query_type = query_type;
  data->cancellable = cancellable;
  data->callback = callback;
  data->user_data = user_data;
  data->destroy_data = destroy_data;

  return data;
}


static void
photos_tracker_queue_log_query (PhotosQuery *query)
{
  PhotosSource *source;
  const gchar *sparql;
  const gchar *tag;

  tag = photos_query_get_tag (query);
  if (tag != NULL && tag[0] != '\0')
    photos_debug (PHOTOS_DEBUG_TRACKER, "%s", tag);

  source = photos_query_get_source (query);
  if (source != NULL)
    {
      const gchar *id;
      const gchar *name;

      id = photos_filterable_get_id (PHOTOS_FILTERABLE (source));
      name = photos_source_get_name (source);
      photos_debug (PHOTOS_DEBUG_TRACKER, "Source (%s): %s", id, name);
    }

  sparql = photos_query_get_sparql (query);
  photos_debug (PHOTOS_DEBUG_TRACKER, "%s", sparql);
}


static void
photos_tracker_queue_collector (GObject *source_object, GAsyncResult *res, gpointer user_data)
{
  g_autoptr (PhotosTrackerQueue) self = PHOTOS_TRACKER_QUEUE (user_data);
  g_autoptr (PhotosTrackerQueueData) data = NULL;

  photos_debug (PHOTOS_DEBUG_TRACKER, "Query processed");

  data = g_queue_pop_head (self->queue);
  if (data->callback != NULL)
    (*data->callback) (source_object, res, data->user_data);
  self->running = FALSE;

  photos_tracker_queue_check (self);
}


static void
photos_tracker_queue_check (PhotosTrackerQueue *self)
{
  PhotosTrackerQueueData *data;
  const gchar *sparql;

  if (self->running)
    return;

  if (self->queue->length == 0)
    return;

  data = g_queue_peek_head (self->queue);
  self->running = TRUE;

  photos_tracker_queue_log_query (data->query);
  sparql = photos_query_get_sparql (data->query);

  switch (data->query_type)
    {
    case PHOTOS_TRACKER_QUERY_SELECT:
      tracker_sparql_connection_query_async (self->local_connection,
                                             sparql,
                                             data->cancellable,
                                             photos_tracker_queue_collector,
                                             g_object_ref (self));
      break;

    case PHOTOS_TRACKER_QUERY_UPDATE:
      tracker_sparql_connection_update_async (self->local_connection,
                                              sparql,
                                              data->cancellable,
                                              photos_tracker_queue_collector,
                                              g_object_ref (self));
      break;

    case PHOTOS_TRACKER_QUERY_UPDATE_BLANK:
      tracker_sparql_connection_update_blank_async (self->local_connection,
                                                    sparql,
                                                    data->cancellable,
                                                    photos_tracker_queue_collector,
                                                    g_object_ref (self));
      break;

    default:
      g_assert_not_reached ();
      break;
    }
}


static gboolean
photos_tracker_queue_start_session_miner_fs (PhotosTrackerQueue *self, GError **error)
{
  const gchar *busname = "org.freedesktop.Tracker3.Miner.Files";
  const gchar *control_busname = "org.freedesktop.Tracker3.Miner.Files.Control";

  photos_debug (PHOTOS_DEBUG_TRACKER, "Connecting to %s", busname);
  self->miner_fs_connection = tracker_sparql_connection_bus_new (busname, NULL, NULL, error);
  if (*error)
    {
      g_warning ("Unable to create connection for session-wide Tracker indexer at %s: %s", busname, (*error)->message);
      return FALSE;
    }

  self->miner_fs_busname = busname;
  self->miner_fs_control_busname = control_busname;

  return TRUE;
}


static gboolean
photos_tracker_queue_start_local_miner_fs (PhotosTrackerQueue *self, GError **error)
{
  const gchar *busname = "org.gnome.Photos.Tracker3.Miner.Files";
  const gchar *control_busname = "org.gnome.Photos.Tracker3.Miner.Files.Control";

  photos_debug (PHOTOS_DEBUG_TRACKER, "Connecting to %s", busname);
  self->miner_fs_connection = tracker_sparql_connection_bus_new (busname, NULL, NULL, error);
  if (*error)
    {
      g_critical ("Could not start local Tracker indexer at %s: %s", busname, (*error)->message);
      return FALSE;
    }

  self->miner_fs_busname = busname;
  self->miner_fs_control_busname = control_busname;

  return TRUE;
}

static gboolean
inside_flatpak (void)
{
    return g_file_test ("/.flatpak-info", G_FILE_TEST_EXISTS);
}

static GObject *
photos_tracker_queue_constructor (GType type, guint n_construct_params, GObjectConstructParam *construct_params)
{
  static GObject *self = NULL;

  if (self == NULL)
    {
      self = G_OBJECT_CLASS (photos_tracker_queue_parent_class)->constructor (type,
                                                                              n_construct_params,
                                                                              construct_params);
      g_object_add_weak_pointer (self, (gpointer) &self);
      return self;
    }

  return g_object_ref (self);
}


static void
photos_tracker_queue_dispose (GObject *object)
{
  PhotosTrackerQueue *self = PHOTOS_TRACKER_QUEUE (object);

  g_clear_object (&self->local_connection);
  g_clear_object (&self->miner_fs_connection);

  G_OBJECT_CLASS (photos_tracker_queue_parent_class)->dispose (object);
}


static void
photos_tracker_queue_finalize (GObject *object)
{
  PhotosTrackerQueue *self = PHOTOS_TRACKER_QUEUE (object);

  g_clear_error (&self->initialization_error);
  g_queue_free (self->queue);

  G_OBJECT_CLASS (photos_tracker_queue_parent_class)->finalize (object);
}


static void
photos_tracker_queue_init (PhotosTrackerQueue *self)
{
  self->queue = g_queue_new ();
}

static void
photos_tracker_queue_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  PhotosTrackerQueue *self = PHOTOS_TRACKER_QUEUE (object);

  switch (prop_id)
    {
    case PROP_MINER_FS_READY:
      g_value_set_boolean (value, self->miner_fs_ready);
      break;

    case PROP_MINER_FS_BUSNAME:
      g_value_set_string (value, self->miner_fs_busname);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
photos_tracker_queue_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  switch (prop_id)
    {
    /* All properties are read only */
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
photos_tracker_queue_class_init (PhotosTrackerQueueClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);

  object_class->constructor = photos_tracker_queue_constructor;
  object_class->dispose = photos_tracker_queue_dispose;
  object_class->finalize = photos_tracker_queue_finalize;
  object_class->get_property = photos_tracker_queue_get_property;
  object_class->set_property = photos_tracker_queue_set_property;

  g_object_class_install_property (object_class,
                                   PROP_MINER_FS_READY,
                                   g_param_spec_boolean ("tracker-miner-fs-ready",
                                                         "Tracker Miner FS ready",
                                                         "TRUE if it is possible to query Tracker indexer",
                                                         FALSE,
                                                         G_PARAM_READABLE));

  g_object_class_install_property (object_class,
                                   PROP_MINER_FS_BUSNAME,
                                   g_param_spec_string ("tracker-miner-fs-busname",
                                                        "Tracker Miner FS busname",
                                                        "D-Bus name of the Tracker indexer daemon",
                                                        "",
                                                        G_PARAM_READABLE));
}


static gboolean
photos_tracker_queue_initable_init (GInitable *initable, GCancellable *cancellable, GError **error)
{
  PhotosTrackerQueue *self = PHOTOS_TRACKER_QUEUE (initable);
  TrackerSparqlConnectionFlags tracker_flags;
  g_autoptr (GFile) store = NULL;
  gboolean ret_val = FALSE;
  gboolean miner_ok = FALSE;

  G_LOCK (init_lock);

  if (self->is_initialized)
    {
      if (self->local_connection != NULL && self->miner_fs_connection != NULL)
        ret_val = TRUE;
      else
        g_assert_nonnull (self->initialization_error);

      goto out;
    }

  g_assert_no_error (self->initialization_error);

  /* Connect to the local database which stores user data.
   *
   * Same flags that tracker-miner-fs uses by default. See:
   * https://gitlab.gnome.org/GNOME/tracker-miners/-/blob/master/src/miners/fs/tracker-main.c#L735 and
   * https://gitlab.gnome.org/GNOME/tracker-miners/-/blob/master/data/org.freedesktop.Tracker.FTS.gschema.xml */
  tracker_flags = TRACKER_SPARQL_CONNECTION_FLAGS_FTS_ENABLE_UNACCENT |
                  TRACKER_SPARQL_CONNECTION_FLAGS_FTS_ENABLE_STOP_WORDS |
                  TRACKER_SPARQL_CONNECTION_FLAGS_FTS_IGNORE_NUMBERS;

  store = g_file_new_build_filename (g_get_user_data_dir (), "gnome-photos", NULL);

  photos_debug (PHOTOS_DEBUG_TRACKER, "Opening local database at %s", g_file_peek_path (store));
  self->local_connection = tracker_sparql_connection_new (tracker_flags,
                                                          store,
                                                          tracker_sparql_get_ontology_nepomuk (),
                                                          cancellable,
                                                          &self->initialization_error);

  if (G_UNLIKELY (self->initialization_error != NULL))
    goto out;

  /* Connect to filesystem indexer. */
  miner_ok = photos_tracker_queue_start_session_miner_fs (self, &self->initialization_error);

  if (!miner_ok && inside_flatpak ())
    {
      g_clear_error (&self->initialization_error);
      miner_ok = photos_tracker_queue_start_local_miner_fs (self, &self->initialization_error);
    }

  if (miner_ok)
    {
      photos_debug (PHOTOS_DEBUG_TRACKER, "Using %s as tracker-miner-fs", self->miner_fs_busname);
      ret_val = TRUE;
    }
  else
    {
      photos_debug (PHOTOS_DEBUG_TRACKER, "Initialization failed due to %s", self->initialization_error->message);
      g_clear_object (&self->miner_fs_connection);
      g_clear_object (&self->local_connection);
      ret_val = FALSE;
    }


 out:
  self->is_initialized = TRUE;
  if (!ret_val)
    {
      g_assert_nonnull (self->initialization_error);
      g_propagate_error (error, g_error_copy (self->initialization_error));
    }

  G_UNLOCK (init_lock);
  return ret_val;
}


static void
photos_tracker_queue_initable_iface_init (GInitableIface *iface)
{
  iface->init = photos_tracker_queue_initable_init;
}


PhotosTrackerQueue *
photos_tracker_queue_dup_singleton (GCancellable *cancellable, GError **error)
{
  GObject *singleton;

  g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  singleton = g_object_new (PHOTOS_TYPE_TRACKER_QUEUE, NULL);

  if (g_initable_init (G_INITABLE (singleton), cancellable, error))
      return PHOTOS_TRACKER_QUEUE (singleton);

  /* On error we deliberately don't unref the object so that we won't try
   * and re-initialize Tracker when called again.
   */
  return NULL;
}


const gchar *
photos_tracker_queue_get_miner_fs_busname (PhotosTrackerQueue *self)
{
  return self->miner_fs_busname;
}


const gchar *
photos_tracker_queue_get_miner_fs_control_busname (PhotosTrackerQueue *self)
{
  return self->miner_fs_control_busname;
}


TrackerNotifier *
photos_tracker_queue_get_notifier (PhotosTrackerQueue *self)
{
  /* We want notifications on filesystem changes, we don't need them for the
   * local database which we manage ourselves.
   */
  return tracker_sparql_connection_create_notifier (self->miner_fs_connection);
}


void
photos_tracker_queue_select (PhotosTrackerQueue *self,
                             PhotosQuery *query,
                             GCancellable *cancellable,
                             GAsyncReadyCallback callback,
                             gpointer user_data,
                             GDestroyNotify destroy_data)
{
  PhotosTrackerQueueData *data;

  if (cancellable != NULL)
    g_object_ref (cancellable);

  data = photos_tracker_queue_data_new (query,
                                        PHOTOS_TRACKER_QUERY_SELECT,
                                        cancellable,
                                        callback,
                                        user_data,
                                        destroy_data);

  g_queue_push_tail (self->queue, data);
  photos_tracker_queue_check (self);
}



void
photos_tracker_queue_update (PhotosTrackerQueue *self,
                             PhotosQuery *query,
                             GCancellable *cancellable,
                             GAsyncReadyCallback callback,
                             gpointer user_data,
                             GDestroyNotify destroy_data)
{
  PhotosTrackerQueueData *data;

  if (cancellable != NULL)
    g_object_ref (cancellable);

  data = photos_tracker_queue_data_new (query,
                                        PHOTOS_TRACKER_QUERY_UPDATE,
                                        cancellable,
                                        callback,
                                        user_data,
                                        destroy_data);

  g_queue_push_tail (self->queue, data);
  photos_tracker_queue_check (self);
}


void
photos_tracker_queue_update_blank (PhotosTrackerQueue *self,
                                   PhotosQuery *query,
                                   GCancellable *cancellable,
                                   GAsyncReadyCallback callback,
                                   gpointer user_data,
                                   GDestroyNotify destroy_data)
{
  PhotosTrackerQueueData *data;

  if (cancellable != NULL)
    g_object_ref (cancellable);

  data = photos_tracker_queue_data_new (query,
                                        PHOTOS_TRACKER_QUERY_UPDATE_BLANK,
                                        cancellable,
                                        callback,
                                        user_data,
                                        destroy_data);

  g_queue_push_tail (self->queue, data);
  photos_tracker_queue_check (self);
}
